import logging

import numpy as np

logger = logging.getLogger(__name__)

# parameters for later use
NUM_PROCESSES_PRED = 1  # use parallel if NUM_PROCESSES > 1 in prediction step
NUM_THREADS_PRED = 1  # Run multithreading instead of multiprocessing in prediction step
NUM_PROCESSES_AXSI = 1  # use parallel if NUM_PROCESSES > 1 in axsi step
NUM_THREADS_AXSI = 1  # Run multithreading instead of multiprocessing in axsi step
LINEAR_LSQ_METHOD = 'R-quadprog'  # Method for linear least squares, implemented using the 'R-quadprog' package
NON_LINEAR_LSQ_METHOD = 'R-minpack'  # Method for non-linear least squares, implemented using the 'R-minpack' package
ADD_VALS = np.arange(0.1, 32, 0.2)  # [0.1, 0.3, 0.5, ... , 31.9]
N_SPEC = len(ADD_VALS)  # N_SPEC == 160

DEBUG_MODE = False
SUBJ_FOLDER = ''

# least squares parameters (for calc_axsi.py file)
X0 = np.asarray([0.5, 5000.0])
MIN_VAL = np.zeros(2)
MAX_VAL = np.array([1.0, 20000.0])

# R_SERVER_PORT=6311 # not in use - R server is very slow


import os
import subprocess
import sys

from setuptools import setup, find_packages
from setuptools.command.install import install


class CustomInstallCommand(install):
    """Custom installation command to check R and install R packages."""

    def run(self):
        # Check if R is installed
        try:
            script_path = os.path.join(os.path.dirname(__file__), 'AxSI', 'check_r.py')
            process = subprocess.Popen(["python", script_path], stdout=sys.stdout, stderr=sys.stderr, text=True)
            process.communicate()
            if process.returncode != 0:
                print(f"Error: Process returned non-zero exit code {process.returncode}")
                sys.exit(process.returncode)
        except Exception as e:
            raise Exception(e)
        install.run(self)


with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"), "r") as fh:
    long_description = fh.read()

requires = [
    'numpy==1.26.4',
    'scipy==1.14.1',
    'scikit-learn==1.5.2',
    'psutil==6.1.1',
    'nibabel==5.3.0',
    'dipy==1.9.0',
    'lsqAxSI',  # Hila Gast's LSQ function
    'rpy2==3.5.17',
    'ecos==2.0.14',
    'cvxpy==1.5.3',
    'gurobipy==12.0.0',
    'dash==2.18.2',
    'plotly==5.24.1',
    'pandas==2.2.3'
]

setup(
    # required
    name="AxSI",
    version="0.0.3",
    # optional
    author="Hila Gast",
    author_email="",
    description="AxSI for MRI data",
    long_description=long_description,  # This is shown on the package detail package on the Pypi
    long_description_content_type="text/markdown",  # text/plain, text/x-rst (for reStructuredText), text/markdown
    url="",  # github

    classifiers=[  # Meta-data. Examples: https://pypi.org/classifiers/
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=requires,  # Install dependencies
    scripts=[
        'scripts/AxSI_main.py',
        'scripts/nifti_viewer.py',
    ],

    packages=find_packages(),

    data_files=[('docs', ['docs/usage.rst']), ],

    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.rst"],
    },
    cmdclass={
        'install': CustomInstallCommand,  # Use the custom install command for R packages
    },
    #
    # # For sdist only:
    # # True: include only files from MANIFEST.in file (not from package_data).
    # # False: include files from MANIFEST.in file AND from package_data
    # include_package_data=False,
    #
    # tests_require=requires + ['nose2'], # Install requirements when you run: python setup.py test
    # test_suite='nose2.collector.collector',
)



# This file is for computing Prediction decays for AxSI
# predict CSF and predicit_hindered are computed at once in advance
# predict restricted takes to much memory so prepare to compute for each voxel later
import logging
import time
from multiprocessing import Manager

import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import norm
from sklearn.mixture import GaussianMixture

from AxSI import config
from AxSI.ax_dti import DTI_Values
from AxSI.matrix_operations import exp_vec_mat_multi
from AxSI.mri_scan import Scan
from AxSI.toolbox import cart2sph_env, parallel_batches

logger = logging.getLogger(__name__)


class Predictions():
    """
    Main class for predictions
    """

    def __init__(self, scan: Scan, dti1000: DTI_Values) -> None:
        """ Use scan for init, don't keep it """
        self.num_of_vols = scan.get_num_of_voxels()  # number of brain voxels in mask
        # init empty arrays for predict_hindered and predict_csf
        if config.NUM_PROCESSES_PRED == 1:  # in multiprocessing init it with shared memory in predict_hindered() method
            self.md0 = np.zeros(self.num_of_vols)
        self.hindered = np.zeros((self.num_of_vols, scan.get_bval_length()))
        self.csf = np.zeros(self.num_of_vols)
        # keep only necessary data from scan and DTI
        self.grad_dirs = scan.get_grad_dirs()  # used for predict_hindered
        self.max_bval = scan.get_max_bval()  # used for predict_hindered
        self.bshell = scan.get_shell()  # used for MD0 curve fit
        self.params = scan.get_params()  # used for predict_restricted init
        self.signal_0 = scan.get_signal_0()  # used for predict_restricted
        self.eigvec1000 = dti1000.get("eigvec")  # used for predict_restricted_init
        self.dt1000_maps = dti1000.get("dt")  # used for predict_hindered

    def get_hindered_slice(self, i) -> np.ndarray:
        """ Get the ith entry of precomputed decay """
        return self.hindered[i, :]

    def get_restricted_slice(self, i) -> np.ndarray:
        """ not precomputed, run predict_restricted """
        return self.predict_restricted(i)

    def get_csf_slice(self, i) -> float:
        ''' get the ith entry of precomputed decay '''
        return self.csf[i]

    def predict_hindered(self, bshell: np.ndarray, shell_dti: list[DTI_Values]) -> None:
        """
        Main function to compute MD0 and predict hindered for each voxel.
        Parallellism depends on hardcoded value of num_of_processed.

        :param bshell:      bvalues of the shell
        :param shell_dti:   dti values of the shell
        """
        # sMDi[j,k] is the biggest eigval of voxel j under the k'th DTI, dims: (n_vols, len(shell_dti))
        sMDi = np.stack([dti.get("eigval")[:, -1] for dti in shell_dti], axis=-1)  # stack on the columns
        if config.NUM_PROCESSES_PRED > 1:
            with Manager() as manager:
                self.md0 = manager.list([None] * self.num_of_vols)  # Create a shared list

                parallel_batches(self.num_of_vols, config.NUM_PROCESSES_PRED, config.NUM_THREADS_PRED,
                                 self.predict_md0_for_voxel, bshell, sMDi)

                # Copy values from the shared object to self
                self.md0 = np.array(self.md0)
        elif config.NUM_THREADS_PRED > 1:
            parallel_batches(self.num_of_vols, config.NUM_PROCESSES_PRED, config.NUM_THREADS_PRED,
                             self.predict_md0_for_voxel, bshell, sMDi)
        else:
            # calculate MD0 in sequence
            for i in range(self.num_of_vols):
                self.predict_md0_for_voxel(bshell, sMDi, i)

        # calculate hindered decay altogether. The decay values are stored in self.hindered.
        self.predict_hindered_decay(sMDi)

    def predict_hindered_decay(self, sMDi: np.ndarray) -> None:
        """
        Calculate hindered decay for all voxels at once.
        Divide MD0 with eigvals of shell DTI with bvalue=1

        :param sMDi: sMDi[j, k] is the biggest eigval of voxel j under the k'th DTI, dims: (n_vols, len(shell_dti))
        """
        every_fac = self.md0 / sMDi[:, 0]  # sMDi[:, 0] is max eignval for bvalue=1
        every_fac = every_fac.reshape((self.num_of_vols, 1, 1))  # adjust to DT1000_maps dims
        every_d_mat = every_fac * 1000 * self.dt1000_maps  # dims: (n_vols, 3, 3)
        # a series of matrix multiplications
        # self.grad_dirs.shape = (n_frames, 3), every_d_mat.shape=(n_vols, 3, 3), decay.shape=(num_of_vols, n_frames)
        decay = exp_vec_mat_multi(self.grad_dirs, every_d_mat, -self.max_bval, axis=1)
        self.hindered = decay  # set result

    def predict_md0_for_voxel(self, bshell: np.ndarray, smdi: np.ndarray, i: int) -> None:
        """
        For a specific voxel, Use `curve_fit` to model eigvals as a function of `bshell` values,
        where eigvals are the maximum eigenvalue for each DTI with respective bshell.

        :param smdi: biggest eigval of voxel i for each bshell. dims: (len(shell_dti), )
        :return:       first parameter of the model
        """
        fit, covariance = curve_fit(exp1, bshell, smdi[i, :], method='trf', loss='soft_l1')
        fit_a = fit[0]  # there are 2 parameters in exp1.
        self.md0[i] = fit_a

    def init_predict_restricted(self) -> None:
        """
        Calculate the parallel and perpendicular sq values for each voxel and timeframe.
        Save the results in self.q_par_sq_arr and self.q_perp_sq_arr.

        Cannot calculate restricted decay in advance (memory limitations).
        Compute partially and avoid repeated calculations
        """
        factor_angle_term_par = self.calc_factor_angle()  # dims: (n_vols, n_frames)
        factor_angle_term_perp = np.sqrt(1 - factor_angle_term_par ** 2)  # dims: (n_vols, n_frames)
        self.q_par_sq_arr = self.calc_sq(factor_angle_term_par)  # dims: (n_vols, n_frames)
        self.q_perp_sq_arr = self.calc_sq(factor_angle_term_perp)  # dims: (n_vols, n_frames)

    def predict_restricted(self, i: int) -> np.ndarray:
        """
        Predict restricted decay for the ith voxel for all timeframes.
        Use data computed to all voxels.
        output dims:

        :param i:  index of the voxel
        :return:   restricted decay values: (n_frames, N_SPEC)
        """
        signal_0 = self.signal_0[i]  # mean of voxel i signal with bvalue=0
        q_par_sq_arr = self.q_par_sq_arr[i, :]  # a value for each timeframe
        q_perp_sq_arr = self.q_perp_sq_arr[i, :]
        md = self.md0[i]  # MD0 from predict_hindered calculations (float).

        E = self.calc_E(q_perp_sq_arr, q_par_sq_arr, md)  # calculate E matrix, dims: (n_frames, N_SPEC)
        decay = signal_0 * E  # consider signal
        self.restricted = decay / np.nanmax(decay)  # normalize with max value

        return self.restricted

    def calc_factor_angle(self) -> np.ndarray:
        """
        Compute for all eigvec at once:
        dim eigvecs: (n_vols, ), dim factor_angle: (n_vols, n_frames)

        :return: factor_angle for each voxel and timeframe. dims: (n_vols, n_frames)
        """
        theta_q = self.params['theta']  # dims: (n_frames, )
        phi_q = self.params['phi']  # dims: (n_frames, )
        _, phi_n, theta_n = cart2sph_env(self.eigvec1000.T)  # dims: (n_vols, )
        tmp_cos = np.outer(np.cos(theta_n), np.cos(theta_q))  # dims: (n_vols, n_frames)
        tmp_sin = np.outer(np.sin(theta_n), np.sin(theta_q))  # dims: (n_vols, n_frames)
        factor_angle = abs(tmp_cos * np.cos(phi_q - phi_n[np.newaxis].T) + tmp_sin)  # dims: (num_of_vols, n_frames)
        return factor_angle

    def calc_sq(self, factor_angle: np.ndarray) -> np.ndarray:
        """
        :param factor_angle: factor_angle: (n_vols, n_frames)
        :return: sq of factor_angle (n_vols, n_frames)
        """
        return (self.params['r'] * factor_angle) ** 2  # self.params['r']: (n_frames,)

    def calc_E(self, q_perp: np.ndarray, q_par: np.ndarray, MD: float) -> np.ndarray:
        """
        Calculate E matrix for predict restricted for one voxel for all timeframes.
        qprep: (n_frames, ), qpar: (n_frames, )

        :param q_perp: perpendicular sq values
        :param q_par:  parallel sq values
        :param MD:     md0 value
        :return:
        """
        bigdel = self.params['big_delta']  # float
        smldel = self.params['small_delta']  # float
        q_par = q_par[np.newaxis].T  # adjust dims to (n_frames, 1)
        # ADD_VALS is defined in toolbox
        tmp = np.multiply.outer(q_perp, (config.ADD_VALS / 2) ** 2)  # dims: (n_frames, N_SPEC)
        tmp = tmp + (q_par * (bigdel - smldel / 3) * MD)  # dims: (n_frames, N_SPEC)
        E = np.exp(-4 * np.pi ** 2 * tmp)  # dims: (n_frames, N_SPEC)
        return E

    def predict_csf(self, shell_dti: list[DTI_Values]):
        """
        Calculate CSF decay for all voxels at once.
        Use gaussian mixture model.

        :param shell_dti:
        :return: decay for voxels with nonzero md value
        """
        # TODO: Refael: the comment is wrong: shell_dti[0] is for bvalue=1 not bvalue=0
        md_non0 = shell_dti[0].get("md")  # dim: (n_vols,). get md values from shellDTI with bvalue=0
        md_non0 = md_non0[md_non0 > 0]  # take only positive values
        # fit data
        gm = self.fit_gaussian_mixture(md_non0)
        p = self.compute_pdf_array(md_non0, gm)  # array of pdf

        self.csf = p[2] / (p[0] + p[1] + p[2])
        self.csf[np.isnan(self.csf)] = 0

    def compute_pdf_array(self, md_nonzero: np.ndarray, gm: GaussianMixture) -> list:
        """
        Probability to get md value (for all voxels with nonzero md) for each gaussian
        from the mixed gaussian model (we use 3 components).

        :param md_nonzero:  md values for all voxels
        :param gm:          fitted gaussian mixture
        :return:            3 probabilities for voxels with nonzero md
        """
        comp_mu = gm.means_  # means of distribution of model
        comp_sigma = gm.covariances_  # covariances of distribution of model
        # compute pdf for each mu and sigma
        p0 = norm.pdf(md_nonzero, comp_mu[0, 0], np.sqrt(comp_sigma[0, 0]))
        p1 = norm.pdf(md_nonzero, comp_mu[1, 0], np.sqrt(comp_sigma[1, 0]))
        p2 = norm.pdf(md_nonzero, comp_mu[2, 0], np.sqrt(comp_sigma[2, 0]))

        return [p0, p1, p2]

    def fit_gaussian_mixture(self, md_nonzero: np.ndarray) -> GaussianMixture:
        """
        Use scipy gaussian mixture fit data.

        :param md_nonzero: md values of the voxels (only nonzero)
        :return:
        """
        md_nonzero = md_nonzero.reshape(-1, 1)
        sigma = np.full((3, 1, 1), 0.3 * np.nanmax(md_nonzero))
        params = {'n_comp': 3,
                  'mu': np.asarray([0.5, 1, 2]).reshape((3, 1)),
                  'weights': np.array([0.7, 0.2, 0.1]),
                  'sigma': sigma,
                  'tol': 1e-9,
                  'max_iter': 10000,
                  'init_params': 'random'
                  }
        gm = GaussianMixture(
            n_components=params['n_comp'], covariance_type='full', tol=params['tol'],
            reg_covar=1e-7, max_iter=params['max_iter'], init_params=params['init_params'],
            weights_init=params['weights'], means_init=params['mu'], precisions_init=params['sigma']
        ).fit(md_nonzero)  # not precisely as in matlab (predict_csf.m)
        return gm


def exp1(x, a, b):
    """ Function for curve fit in MD0 calculation.
    The learned parameters are: a, b
    """
    y = a * np.exp(b * x)
    return y


def predict(scan: Scan, dti1000: DTI_Values, shell_dti: list[DTI_Values]) -> Predictions:
    """
    Calculate predictions decays for each environment

    :param scan:        scan object
    :param dti1000:     DTI_Values object for dti1000
    :param shell_dti:   list of DTI_Values objects for shell_dti
    :return object with decays prediction for AxSI
    """

    decays = Predictions(scan, dti1000)  # init object
    # dims hindered_decay: (n_vols, len(shell_dti)),   dims md0: (n_vols, ). Stored in self.hindered.
    tic = time.time()
    decays.predict_hindered(scan.get_shell(), shell_dti)
    toc = time.time()
    logger.debug("Computation time md0 all voxels " + (
        "with multi-processes/multi-threads " if config.NUM_PROCESSES_PRED > 1 or config.NUM_THREADS_PRED > 1 else "") + "= " + str(
        (toc - tic)) + " sec.")

    # predict restricted cannot be calculated in advance because of memory limitations
    tic = time.time()
    decays.init_predict_restricted()
    toc = time.time()
    logger.debug("Computation time restricted decay = " + str((toc - tic)) + " sec.")

    # calculate CSF decay, dims: (n_vols)
    decays.predict_csf(shell_dti)
    return decays



# This file is for running DTI for AxSI analysis
import logging
from pathlib import Path

import numpy as np

from AxSI.matrix_operations import array_to_sym, sym_mat_to_array
from AxSI.mri_scan import Scan
from AxSI.toolbox import Values, MyManager

logger = logging.getLogger(__name__)

EPS = np.finfo('float').eps  # 2.22X10-16
KEYS = ['dt', 'md', 'fa', 'eigval', 'eigvec']  # attributes of DTI calculations


class DTI_Values(Values):
    """
    An extended class for Values with relevant attributes for DTI
    """

    def __init__(self, n: int) -> None:
        """
        :param n: number of voxels
        """
        self.fa = np.zeros(n)
        self.md = np.zeros(n)
        self.dt = np.zeros((n, 3, 3))
        self.eigval = np.zeros((n, 3))
        self.eigvec = np.zeros((n, 3))


# register class to MyManager, used for multiprocessing
MyManager.register('DTI_Values', DTI_Values)


class AXSI_DTI:
    """
    The main object for performing DTI
    """

    def __init__(self, scan: Scan, bvalue: float, is1000: bool) -> None:
        """
        Use scan for initialization, don't keep it

        :param scan:    scan object
        :param bvalue:  bvalue for the DTI
        :param is1000:
        """
        self.num_of_vols = scan.get_num_of_voxels()  # number of brain voxels in mask
        self.values = DTI_Values(self.num_of_vols)  # init struct with zero values
        # compute in advance
        bvlocs = scan.get_locs_by_bvalues(bvalue)  # indices of specific bvalue
        self.bval_arr = self.compute_bval_arr(scan, bvlocs)  # dims: (len(bvlocs), 6)
        # signal log is computed differently for DTI1000 and shellDTI
        # dims: (n_vols, len(bvlocs)) - n_vols after masking
        self.signal_log = self.compute_signal_log(scan, bvlocs, is1000)

    def main(self) -> DTI_Values:
        """
        Calculate DTI to each voxel separatelly.

        :return: struct with dti values (not a shared one in any case)
        """
        # comment out for now. multiprocessing in this step is not efficient, instead we run each shell
        # with all voxels in parallel.
        if False:  # config.NUM_PROCESSES == 0:
            pass
            # logger.info('Run parallelism with num_of_processes={}'.format(config.NUM_PROCESSES))
            # # run parallelly
            # with MyManager() as manager:
            #     # create a shared object to use in mutliprocessing
            #     self.values = manager.DTI_Values(self.num_of_vols)
            #     parallel_batches(self.num_of_vols, self.volume_dti, False)
            #     self.extract_data()  # copy data from shared object to regular object
        else:
            # run sequentially
            for i in range(self.num_of_vols):
                self.volume_dti(i)
        return self.values

    def volume_dti(self, i: int) -> None:
        """
        Calculate dti measurements for the i'th voxel
        :param i: index of the voxel
        """
        self.volume_dt_calc(i)
        self.volume_eigen_calc(i)
        self.volume_md_calc(i)
        self.volume_fa_calc(i)

    def volume_dt_calc(self, i: int) -> None:
        """
        Calculate diffusion tensor for the i'th voxel over all len(bvlocs).
        Diffusion tensor is a (3,3) symmetric matrix.

        :param i: index of the voxel
        """
        bval_arr = self.bval_arr  # dims: (len(bvlocs), 6)
        signal_log_i = -1 * (self.signal_log[i, :])  # dims: (len(bvlocs), )
        dt_i = np.linalg.lstsq(bval_arr, signal_log_i, rcond=None)[0]  # dims: (6,)
        self.values.set('dt', array_to_sym(dt_i), i)  # dims: (3,3)

    def volume_md_calc(self, i: int) -> None:
        """
        Calculate mean diffusivity for the i'th voxel mean diffusivity is a float

        :param i: index of the voxel
        """
        md_i = np.mean(self.values.get('eigval', i))
        self.values.set('md', md_i, i)

    def volume_fa_calc(self, i: int) -> None:
        """
        Calculate fractional anisotropy for the i'th voxel

        :param i: index of the voxel
        """
        eigval_i = self.values.get('eigval', i)
        md_i = self.values.get('md', i)
        fa_i = np.sqrt(1.5) * np.linalg.norm(eigval_i - md_i) / np.linalg.norm(eigval_i)
        self.values.set('fa', fa_i, i)

    def volume_eigen_calc(self, i: int) -> None:
        """
        Calculate eigen values and eigen vectors of the diffusion tensor of the i'th voxel
        :param i: index of the voxel
        """
        dt_mat = self.values.get('dt', i)  # shape: (3, 3)
        eigen_vals, eigen_vecs = np.linalg.eig(dt_mat)  # compute eigen values and vectors
        index = np.argsort(eigen_vals)  # get indices of sorted eigen values
        eigen_vals = eigen_vals[index] * 1000
        self.volume_eigval_calc(eigen_vals, i)
        self.volume_eigvec_calc(eigen_vecs, eigen_vals, index, i)

    def volume_eigval_calc(self, eigen_vals: np.ndarray, i: int) -> None:
        """
        Change the eigen values to their absolute values (or epsilon) and set them sorted in increasing order in
        attribute self.values.eigval

        :param eigen_vals: sorted eigen values based on eigen values increasing order
        :param i: index of the voxel
        """
        if np.all(eigen_vals < 0):  # if all eigen values are negative:
            eigen_vals = np.abs(eigen_vals)  # change the sign of all values
        eigen_vals[eigen_vals <= 0] = EPS  # replace negative eigen values with EPS
        self.values.set('eigval', eigen_vals, i)

    def volume_eigvec_calc(self, eigen_vecs: np.ndarray, eigen_vals: np.ndarray, index: int, i: int) -> None:
        """
        Sort eigen vectors and keep (in self.values.eigvec attribute) the "last" one based on eigen values increasing
        order (multiplied by its eigen value)

        :param eigen_vecs: eigen vectors
        :param eigen_vals: sorted eigen values based on eigen values increasing order
        :param index: indexes of the eigen values according their sorted values
        :param i: index of the voxel
        """

        eigen_vecs = eigen_vecs[:, index]  # sort based in eigen values order
        eigen_vecs = eigen_vecs[:, -1] * eigen_vals[-1]  # take the last vector and multiply it
        self.values.set('eigvec', eigen_vecs, i)

    def extract_data(self) -> None:
        """
        Copy data from shared object to a new regular object set the new object instead of the shared one
        """
        shared_values = self.values
        output_values = DTI_Values(self.num_of_vols)  # init new object
        DTI_Values.copy_values(shared_values, output_values, KEYS)  # copy from shared to new
        self.values = output_values  # replace it

    def compute_bval_arr(self, scan: Scan, bvlocs: np.ndarray) -> np.ndarray:
        """
        Compute in advance for tensors calculation

        :param scan:    Scan object
        :param bvlocs:  locations of specific bvalue
        :return:        array of shape (len(bvlocs), 6)
        """
        n = len(bvlocs)
        bval_arr = np.zeros((n, 6))
        bval_real = scan.get_bval_data()[bvlocs]  # dims: (len(bvlocs), )
        norm_bvec = scan.get_bvec_norm()[bvlocs, :]  # dims: (len(bvlocs), 3)

        for i in range(n):
            bmat = bval_real[i] * np.outer(norm_bvec[i, :], norm_bvec[i, :])  # dims: (3,3)
            bval_arr[i, :] = sym_mat_to_array(bmat)  # dims: (6, )

        return bval_arr  # dims: (len(bvlocs), 6)

    def compute_signal_log(self, scan: Scan, bvlocs: np.ndarray, is1000: bool) -> np.ndarray:
        """
        :param scan:    Scan object
        :param bvlocs:  locations of specific bvalue
        :param is1000:  If `1`, the signal is divided by the signal at the initial time point where `bvals=0`;
                        otherwise, it is divided by the average of signals where `bvals=0`.
        :return:        matrix of dims: (n_vols, len(bvlocs)) - n_vols after masking
        """
        signal = scan.get_smoothed_data(bvalue_index=bvlocs)  # dims (len(n_vols), len(bvlocs))
        signal_0 = scan.get_signal_0(is1000)[np.newaxis].T  # dims (len(n_vols), 1)

        signal_log = np.log((signal / signal_0) + EPS)  # dims: (len(n_vols), len(bvlocs)))
        signal_log[np.isnan(signal_log)] = 0
        signal_log[np.isinf(signal_log)] = 0

        return signal_log

    def save_dti_files(self, scan: Scan, subj_folder: Path) -> None:
        """
        Save data of each measurement to file
        :param scan:    Scan object
        :param subj_folder: folder where data is saved
        """
        files_dict = self.values.get_dict(KEYS)  # create dict of keys and data
        scan.save_files(files_dict, subj_folder)  # use scan object to save data in dict


def dti(scan: Scan, bvalue: float, is1000: bool = False, subj_folder: Path = None) -> DTI_Values:
    """
    Envelope function for running AxSI DTI of specific bvalue

    :param scan:        scan object
    :param bvalue:      bvalue to calculate envelope for
    :param is1000:      flag to distinguish DTI1000 and shellDTI
    :param subj_folder: data will be saved to files iff path is given
    :return:            object containing data calculated in DTI
    """
    bvalue_dti = AXSI_DTI(scan, bvalue, is1000)  # init calc object
    # call main function of calc object
    values = bvalue_dti.main()  # return value is a "result" object
    if subj_folder:  # if path was given
        bvalue_dti.save_dti_files(scan, subj_folder)  # save files to path
    return values



# The main file for AxSI measures calculations

import logging

import cvxpy as cp
import numpy as np
import rpy2
import rpy2.robjects as ro
from gurobipy import GRB
from rpy2.robjects.numpy2ri import activate
from rpy2.robjects.packages import importr
from scipy.optimize import minimize

activate()  # Enable automatic conversion between R and NumPy

# Import necessary R packages
quadprog = importr("quadprog")

logging.getLogger('rpy2').setLevel(logging.WARNING)  # Suppress DEBUG logs from rpy2

logger = logging.getLogger(__name__)


def lin_least_squares_with_constrains_r(A: np.ndarray, b: np.ndarray, lb: np.ndarray, ub: np.ndarray) -> np.ndarray:
    """
    Solve linear least squares with constraints using R's 'quadprog' package.

    Parameters:
        A (np.ndarray): The coefficient matrix.
        b (np.ndarray): The target vector.
        lb (np.ndarray): Lower bounds on the solution.
        ub (np.ndarray): Upper bounds on the solution.
        start_p (np.ndarray): Starting point for the linear least squares.

    Returns:
        np.ndarray: The optimized solution vector.
    """
    # tic =time.time()
    # Define the D matrix (A^T A) and d vector (-A^T b) for quadprog
    D = A.T @ A
    d = A.T @ b

    D_r = ro.r.matrix(D, nrow=D.shape[0], ncol=D.shape[1])
    d_r = ro.FloatVector(d)

    # Constraints matrix: combine lb and ub as inequality constraints
    # G matrix: [-I; I] (negative and positive identity matrices for bounds)
    G = np.vstack([np.ones((A.shape[1])), np.eye(A.shape[1]), -np.eye(A.shape[1])])
    h = np.hstack([1, lb, -ub])

    G_r = ro.r.matrix(G, nrow=G.shape[0], ncol=G.shape[1])
    h_r = ro.FloatVector(h)
    try:
        # Solve the quadratic programming problem using solve.QP.
        # https://www.rdocumentation.org/packages/quadprog/versions/1.5-8/topics/solve.QP
        result = quadprog.solve_QP(Dmat=D_r, dvec=d_r, Amat=G_r.T, bvec=h_r, meq=1)
        # Extract the solution
        solution = result.rx2("solution")
        return solution
    except rpy2.rinterface_lib.embedded.RRuntimeError:
        logger.info("SolverError: The linear LSQ problem failed.")
        return np.ones(A.shape[1])


def lin_least_squares_with_constraints_gurobi(A: np.ndarray, b: np.ndarray, lb: np.ndarray,
                                              ub: np.ndarray, model) -> np.ndarray:
    # Gurobi is not thread-safe, so a separate environment is required for each process/thread when using
    # multiprocessing or multithreading.
    # However, it's not necessary for multiprocessing in our case since we are utilizing Gurobi's internal parallelism
    # rather than Python's multiprocessing.
    # As for multithreading, it appears to work fine without a separate environment, but we include it as a precaution.
    # See here for more details:
    # https://support.gurobi.com/hc/en-us/articles/360043111231-How-do-I-use-multiprocessing-in-Python-with-Gurobi

    model.setParam('OutputFlag', 0)
    # Cannot run with Python's multiprocessing when using more than 1 thread in Gurobi's internal multiprocessing
    # (Gurobi's "threads" operate as separate processes). Python threads can coexist with Gurobi's threads,
    # but we always set Gurobi's thread count to 1 to maintain control over CPU usage.
    model.setParam("Threads", 1)

    # Create decision variables x (continuous variables between lb and ub)
    x = model.addMVar(A.shape[1], lb=lb, ub=ub, name="x")  # Multi-dimensional variables

    # Objective: Minimize the quadratic term 0.5 * (A @ x - b)^2
    error = A @ x - b
    quad_expr = 0.5 * (error @ error)  # (A @ x - b)^T @ (A @ x - b)

    # Set the objective to minimize the quadratic expression
    model.setObjective(quad_expr, GRB.MINIMIZE)

    # Add the sum(x) == 1 constraint
    model.addConstr(x.sum() == 1)
    # Optimize the model
    model.optimize()
    # Check the optimization result
    if model.status == GRB.OPTIMAL:
        return x.X
    else:
        logger.info("SolverError: The linear LSQ problem failed.")
        return np.ones(x.shape)


def lin_least_squares_with_constraints_scipy(A: np.ndarray, b: np.ndarray, lb: np.ndarray,
                                             ub: np.ndarray) -> np.ndarray:
    """
    use scipy.minimize to find x such that :
        minimize sum of squares for [A @ x - b]
        lb <= x <= ub and sum(x) == 1

    Parameters
    ----------
    A: Xprim
    b: Yprim
    lb: LOWER_BOUND
    ub: UPPER_BOUND

    Returns
    -------

    """

    # Define the objective function (least squares)
    def objective(x):
        return 0.5 * np.sum((A @ x - b) ** 2)

    def constraint_sum(x):
        return np.sum(x) - 1  # sum(x) == 1

    # Define constraints
    constraints = [{'type': 'eq', 'fun': constraint_sum}]
    x0 = np.zeros(A.shape[1])

    # Solve using minimize with bounds
    # x = minimize(objective, x0, constraints=constraints, bounds=np.column_stack((lb, ub)), method='SLSQP', options={'ftol': 1e-6, 'disp': True}).x
    # x = minimize(objective, x0, constraints=constraints, bounds=np.column_stack((lb, ub)), method='trust-constr').x
    x = minimize(objective, x0, constraints=constraints, bounds=np.column_stack((lb, ub))).x
    return x


def lin_least_squares_with_constraints_cvxpy(A: np.ndarray, b: np.ndarray, lb: np.ndarray,
                                             ub: np.ndarray) -> np.ndarray:
    """
    use cvxpy to find x such that :
    minimize sum of squares for [A @ x - b]
    lb <= x <= ub and sum(x) == 1
    """
    # Define your variables and problem data
    n = A.shape[1]
    x = cp.Variable(n)
    objective = cp.Minimize(0.5 * cp.sum_squares(A @ x - b))
    constraints = [x >= lb, x <= ub, cp.sum(x) == 1]

    # Create the optimization problem
    prob = cp.Problem(objective, constraints)

    # Try to solve the problem
    try:
        prob.solve(warm_start=True, solver=cp.ECOS)
        # Optionally, you can check the status and the optimal value:
    except cp.SolverError:
        logger.info("SolverError: The linear LSQ problem does not have a feasible solution.")
        return np.ones(x.shape)

    return x.value



# This file provide methods for matrix calculations

import logging
import warnings

import numpy as np

logger = logging.getLogger(__name__)


def array_to_sym(arr: np.ndarray) -> np.ndarray:
    """
    Converts an nd-array into a symmetric matrix.
    Expect len(arr) to be number of element in upper triangle of a squared matrix
    example for input: [1,2,3,4,5,6]:
        1, 2, 3
        2, 4, 5
        3, 5, 6

    :param arr: 1D array
    :return: 2D matrix of shape (size, size)
    """
    size = get_mat_size_by_n_elements(len(arr))
    mat = np.zeros([size, size])
    mat[np.triu_indices(size)] = arr
    mat = mat.T
    mat[np.triu_indices(size)] = arr
    return mat


def get_mat_size_by_n_elements(n: int) -> int:
    """
    return the size of squared matrix
    with n elements in upper triangle
    print message and return n if no such solution
    """
    roots = np.roots([1, 1, -2 * n])
    pos_roots = roots[roots > 0]
    if pos_roots[0].is_integer():
        return int(pos_roots[0])
    else:
        logger.info(f"could not convert array of len: {n} to symmetric matrix")
        return n


def sym_mat_to_array(mat: np.ndarray) -> np.ndarray:
    """
    1D array with matrix elements. elements with above diagonal are multiplied by 2
    for input: 1, 2, 3
               2, 4, 5
               3, 5, 6
    output:    [1, 2*2, 2*3, 4, 2*5, 6]
    :param mat: symmetric matrix
    :return:    1D array with matrix elements
    """
    mat = mat + np.triu(mat, k=1)  # multiply by two elements above diag
    arr = mat[np.triu_indices(3)]  # build array with elements of upper triangular
    return arr


def normalize_matrix_rows(mat: np.ndarray) -> np.ndarray:
    """
    Normalize each row in input separately

    :param mat: 2D matrix
    :return: normalized matrix according to rows
    """
    rows_norm = np.linalg.norm(mat, axis=1, keepdims=True)  # compute array of norm values of each row
    # divide each row by its own norm, ignoring zero rows
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        norm_mat = np.where(rows_norm > 0, mat / rows_norm, 0)
    return norm_mat


def exp_vec_mat_multi(vec: np.ndarray, mat: np.ndarray, coeff: float, axis=0) -> np.ndarray:
    """
    A series of matrix and vector multiplications and exponents

    :param vec:     vectors. can be several vectors (a matrix)
    :param mat:     matrix of shape ([any], vector.shape[axis], vector.shape[axis]), where axis is the parameter axis.
                    The first dimension is optional, if exists, broadcasting will be done on this dimension.
    :param coeff:   float. coefficient of vector multiplication
    :param axis:    how the vectors are stored in the vec. 0 - the vector are in rows of vec, 1 - in columns of vec.
    :return:        e^(diag(coeff * vec @ mat @ trans(vec)))
                    where diag is function that take only vector with the diagonal elements.
    """
    multi_mat_vec = np.dot(mat, vec.T)  # mat @ trans(vec)
    # array of main diagonal of vec @ mat @ trans(vec)
    multi_vec_mat_vec = np.sum(multi_mat_vec * vec.T, axis=axis)
    exp = np.exp(coeff * multi_vec_mat_vec)
    return exp



# The main file for AxSI measures calculations

import logging
import time
from pathlib import Path

import gurobipy as gb
import numpy as np
from lsqAxSI.lsq_axsi import nonlinear_least_squares
from scipy.optimize import least_squares

from AxSI import config
from AxSI.ax_dti import DTI_Values
from AxSI.linear_lsq import lin_least_squares_with_constrains_r, lin_least_squares_with_constraints_gurobi, \
    lin_least_squares_with_constraints_scipy, lin_least_squares_with_constraints_cvxpy
from AxSI.matrix_operations import exp_vec_mat_multi
from AxSI.mri_scan import Scan
from AxSI.non_linear_lsq import least_squares_envelope_r, reg_func_r, jac_calc_r, reg_func, jac_calc
from AxSI.predictions import Predictions
from AxSI.predictions import predict
from AxSI.toolbox import MyManager, Values, parallel_batches
from AxSI.toolbox import init_l_matrix, init_yd

logger = logging.getLogger(__name__)

# attributes of DTI calculations
KEYS = ['pfr', 'ph', 'pcsf', 'pasi', 'paxsi', 'CMDfr', 'CMDfh']


class Calc_AxSI_Values(Values):
    """An extended class for Values with relevant attributes for AxSI """

    def __init__(self, n: int) -> None:
        """
        Parameters
        ----------
        n: number of voxels
        """
        self.pfr = np.zeros(n)
        self.ph = np.zeros(n)
        self.pcsf = np.zeros(n)
        self.pasi = np.zeros(n)
        self.paxsi = np.zeros((n, config.N_SPEC))  # N_SPEC is defined in toolbox
        self.CMDfr = np.zeros(n)
        self.CMDfh = np.zeros((n, 2))


# register class to MyManager, used for multiprocessing
MyManager.register('Calc_AxSI_Values', Calc_AxSI_Values)


class Calc_AxSI():
    """ The main object for performing AxSI analysis """

    def __init__(self, scan: Scan, decays: Predictions) -> None:
        """
        Use Scan for init, don't keep it

        :param scan:    scan object
        :param decays:  object with the decays of hindered, restricted and csf.
        """
        self.YD = init_yd()
        self.L_MATRIX = init_l_matrix(config.N_SPEC)  # dim: (N_SPEC, N_SPEC+2)
        self.lower_bound = np.zeros(config.N_SPEC + 1)
        self.upper_bound = np.ones(config.N_SPEC + 1)
        self.decays = decays  # Predictions object
        self.data = scan.get_smoothed_data()  # smoothed MRI signal
        self.n_vols = scan.get_num_of_voxels()  # number of brain voxels in mask
        # in multiprocessing the shared memory initialized in main() method.
        if config.NUM_PROCESSES_AXSI == 1:
            self.values = Calc_AxSI_Values(self.n_vols)  # init struct with zero values
        # compute in advance
        self.vCSF = init_pixpredictCSF(scan.get_grad_dirs())  # depends on scan values, dims: (n_frames, )

    # The same as in DTI file
    def save_calc_files(self, scan: Scan, subj_folder: Path) -> None:
        """ save data of each measurement to file """
        files_dict = self.values.get_dict(KEYS)  # create dict of keys and data
        scan.save_files(files_dict, subj_folder)  # use scan object to save data in dict

    def main(self) -> None:
        """
        Calculate AxSI to each voxel separatelly
        Parallellism depends on hardcoded value of num_of_processed
        Save a struct with AxSI values (not a shared one in any case)
        """
        if config.NUM_PROCESSES_AXSI > 1:
            # run in parallel
            with MyManager() as manager:
                # Create a shared object
                self.values = manager.Calc_AxSI_Values(self.n_vols)
                parallel_batches(self.n_vols, config.NUM_PROCESSES_AXSI, config.NUM_THREADS_AXSI,
                                 self.perform_iteration, self.decays, self.data, self.vCSF)
                temp = self.values
                self.values = Calc_AxSI_Values(self.n_vols)
                # Copy values from the shared object to self
                Values.copy_values(temp, self.values, KEYS)
        elif config.NUM_THREADS_AXSI > 1:
            parallel_batches(self.n_vols, config.NUM_PROCESSES_AXSI, config.NUM_THREADS_AXSI, self.perform_iteration,
                             self.decays, self.data, self.vCSF)
        else:
            for i in range(self.n_vols):
                self.perform_iteration(self.decays, self.data, self.vCSF, None, i)

        self.values.set('pcsf', self.decays.csf)

    from typing import Any

    def perform_iteration(self, decays: Predictions, ydata: np.ndarray, vCSF: np.ndarray, env: Any, i: int) -> None:
        """
        Calculate AxSI measurements for the i'th voxel.

        :param decays:  Predictions object
        :param ydata:   Smoothed MRI signal (n_vols, n_frames)
        :param vCSF:    Voxel-wise CSF values (n_frames,)
        :param env:     Gurobi environment (could be specific to a process/thread) or None for other packages
        :param i:       Index of the voxel to process
        """
        if i // 1000 == i / 1000:
            logger.info(f'calc axsi iter: {i}')

        # Get data for current voxel
        ydata_i = ydata[i]  # scan signal, dims: (n_frames, )
        vH = decays.get_hindered_slice(i)  # dims: (n_frames, )
        vR = decays.get_restricted_slice(i)  # dims: (n_frames, N_SPEC)
        prcsf_i = decays.get_csf_slice(i)  # float
        vH[vH > 1] = 0  # correct vH data

        # nonlinear least squares with predictions and signal
        parameter_hat = self.least_squares_envelope(ydata_i, vH, vR, vCSF, prcsf_i)
        # linear least squares with predictions and signal
        x = self.solve_vdata(ydata_i, vH, vR, vCSF, prcsf_i, parameter_hat, env)
        # update iteration results in object
        self.set_values(x, parameter_hat, prcsf_i, i)

    def least_squares_envelope(self, ydata_i: np.ndarray, vH: np.ndarray, vR: np.ndarray,
                               vCSF: np.ndarray, prcsf_i: float) -> np.ndarray:
        """
        Use own version to achieve more similar results to MATLAB lsqnonlin

        :param ydata_i: scan signal for current voxel
        :param vH: hindered prediction for current voxel
        :param vR: restricted prediction for current voxel
        :param vCSF: CSF prediction
        :param prcsf_i: CSF fraction for current voxel
        :return: array with two parameters of the model
        """
        vRes = np.dot(vR, self.YD)  # YD is the same for all voxels, dims: (n_frames, )
        vRes = np.nan_to_num(vRes, nan=0)  # replace NaN with 0
        # run nonlinear least squares with regression function
        if config.NON_LINEAR_LSQ_METHOD == 'R-minpack':
            parameter_hat = least_squares_envelope_r(reg_func_r, config.X0, bounds=(config.MIN_VAL, config.MAX_VAL),
                                                     jac=jac_calc_r,
                                                     ftol=1e-6, xtol=1e-6, diff_step=1e-3, max_nfev=20000,
                                                     args=(ydata_i, vH, vRes, vCSF, prcsf_i))
        elif config.NON_LINEAR_LSQ_METHOD == 'scipy':
            parameter_hat = least_squares(reg_func, config.X0, jac=jac_calc, bounds=(config.MIN_VAL, config.MAX_VAL),
                                          method='trf',
                                          ftol=1e-6, xtol=1e-6, gtol=1e-06, x_scale=1.0, loss='linear', f_scale=1.0,
                                          diff_step=1e-3, tr_solver=None, tr_options={}, jac_sparsity=None,
                                          max_nfev=20000, verbose=0,
                                          args=(ydata_i, vH, vRes, vCSF, prcsf_i), kwargs={}).x
        elif config.NON_LINEAR_LSQ_METHOD == 'lsq-axsi':
            parameter_hat = nonlinear_least_squares(reg_func, config.X0, bounds=(config.MIN_VAL, config.MAX_VAL),
                                                    jac=jac_calc, ftol=1e-6,
                                                    xtol=1e-6, diff_step=1e-3, max_nfev=20000,
                                                    args=(ydata_i, vH, vRes, vCSF, prcsf_i)).x
        return parameter_hat

    def solve_vdata(self, ydata_i: np.ndarray, vH: np.ndarray, vR: np.ndarray,
                    vCSF: np.ndarray, prcsf_i: float, parameter_hat: np.ndarray, env=None) -> np.ndarray:
        """
        Linear least squares with predictions and signal

        :param ydata_i: scan signal for current voxel
        :param vH: hindered prediction for current voxel
        :param vR: restricted prediction for current voxel
        :param vCSF: CSF prediction
        :param prcsf_i: CSF fraction for current voxel
        :param parameter_hat: results of lsqnonlin
        :return: np array of shape (N_SPEC+2,)
        """
        # divide signal with nonlinlsq result
        vdata = ydata_i / parameter_hat[1]  # dims: (n_frames, )
        # adjust variables to current voxels
        lower_bound_i = np.append(self.lower_bound, prcsf_i - 0.02)  # add a last element
        upper_bound_i = np.append(self.upper_bound, prcsf_i + 0.02)  # add a last element

        # Stack predictions
        vR = np.nan_to_num(vR, nan=0)  # replace NaN with 0
        preds = np.column_stack((vR, vH, vCSF))  # dims: (n_frames, N_SPEC+2)

        Xprim = np.concatenate((preds, self.L_MATRIX))  # dims: (n_frames+N_SPEC, N_SPEC+2)
        Yprim = np.concatenate((vdata, np.zeros(config.N_SPEC)))  # dims: (n_frames+N_SPEC, )
        # run linear least squares
        if config.LINEAR_LSQ_METHOD == 'R-quadprog':
            x = lin_least_squares_with_constrains_r(Xprim, Yprim, lower_bound_i, upper_bound_i)
        elif config.LINEAR_LSQ_METHOD == 'gurobi':
            with gb.Model(env=env) as model:
                x = lin_least_squares_with_constraints_gurobi(Xprim, Yprim, lower_bound_i, upper_bound_i, model)
        elif config.LINEAR_LSQ_METHOD == 'scipy':
            x = lin_least_squares_with_constraints_scipy(Xprim, Yprim, lower_bound_i, upper_bound_i)
        elif config.LINEAR_LSQ_METHOD == 'cvxpy':
            x = lin_least_squares_with_constraints_cvxpy(Xprim, Yprim, lower_bound_i, upper_bound_i)
        return x  # dims: (N_SPEC+2, )

    def set_values(self, x: np.ndarray, parameter_hat: np.ndarray, prcsf_i: float, i: int) -> None:
        """ update object with current voxel values """
        x[x < 0] = 0  # replace negative values with zeros
        nx = x[:130]  # take the first 130 element
        nx = nx / np.sum(nx)  # normalize
        pasi_i = np.sum(nx * config.ADD_VALS[:130])  # ADD_VALS is defined in toolbox
        self.values.set('ph', x[160], i)
        self.values.set('pfr', np.min([1 - prcsf_i - x[160], 0.]), i)
        self.values.set('pasi', pasi_i, i)
        self.values.set('paxsi', x[:160], i)
        self.values.set('CMDfh', parameter_hat, i)
        self.values.set('CMDfr', 1 - parameter_hat[0] - prcsf_i, i)


def calc_axsi(scan: Scan, dti1000: DTI_Values, shell_dti: list[DTI_Values], subj_folder: Path) -> None:
    """
    Envelope function for running AxSI.
    Use DTI1000 and shellDTI calculated for scan.

    :param scan:        scan object
    :param dti1000:     DTI_Values object for dti1000
    :param shell_dti:   list of DTI_Values objects for shell_dti
    :param subj_folder: data will be saved to files in path
    """
    tic = time.time()
    decays = predict(scan, dti1000, shell_dti)  # Calculate predictions decays for each environment
    toc = time.time()
    logger.debug("Computation time predictions = " + str((toc - tic)) + " sec.")
    # init AxSI object
    calc = Calc_AxSI(scan, decays)
    # run main function of AxSI analysis
    tic = time.time()
    calc.main()
    toc = time.time()
    logger.debug("Computation time Calc_AxSI.main " + (
        "with multi-processes/multi-threads " if config.NUM_PROCESSES_AXSI > 1 or config.NUM_THREADS_AXSI > 1 else "") + "= " + str(
        (toc - tic)) + " sec.")
    logger.info("Saving calc files")
    calc.save_calc_files(scan, subj_folder)


def init_pixpredictCSF(grad_dirs: np.ndarray) -> np.ndarray:
    """
    A series of matrix multiplications

    :param grad_dirs: matrix of (n_frames, 3)
    :return: np array of shape (n_frames,)
    """
    D_mat = np.eye(grad_dirs.shape[1]) * 4  # dim: (3,3)
    pixpredictCSF = exp_vec_mat_multi(grad_dirs, D_mat, -4)
    return pixpredictCSF



import os
import subprocess
import sys


def install_r_package(package_name):
    """Install an R package using the system R."""
    try:
        # Check if the package is already installed by running R's package test
        result = subprocess.run(
            ["R", "-e", f"if (!require('{package_name}', character.only = TRUE)) quit(status = 1)"],
            capture_output=True, text=True
        )

        if result.returncode != 0:
            print(f"Package {package_name} not found. Installing...")
            # Set CRAN mirror and install the package
            install_result = subprocess.run(
                ["R", "-e", f"install.packages('{package_name}', repos='https://cran.r-project.org')"],
                capture_output=True, text=True
            )

            if install_result.returncode != 0:
                print(f"Error installing package {package_name}.")
                print(f"Output: {install_result.stdout}")
                print(f"Error: {install_result.stderr}")
                sys.exit(1)
            else:
                print(f"Package {package_name} successfully installed.")
        else:
            print(f"R package {package_name} is already installed.")
    except subprocess.CalledProcessError as e:
        print(f"Error during package installation: {e}")
        sys.exit(1)
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        sys.exit(1)


def is_r_installed():
    """Check if R is installed and accessible in PATH."""
    try:
        subprocess.run(["R", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        return True
    except FileNotFoundError:
        return False


def get_os_specific_instructions():
    """Provide instructions for adding R to PATH based on the operating system."""
    os_name = os.name
    if os_name == "nt":  # Windows
        instructions = (
            "1. Locate the folder where R is installed (e.g., C:\\Program Files\\R\\R-x.x.x\\bin).\n"
            "2. Add this folder to the PATH environment variable:\n"
            "   - Search for 'Environment Variables' in the Start Menu.\n"
            "   - Click 'Environment Variables'.\n"
            "   - Edit the 'Path' variable and add the R folder path.\n"
            "3. Restart your terminal or IDE."
        )
    elif os_name == "posix":  # macOS or Linux
        instructions = (
            "1. Locate the folder where R is installed (use `which R` to find it, e.g., /usr/local/bin/R).\n"
            "2. Add this folder to the PATH environment variable:\n"
            "   - Edit your shell configuration file (e.g., ~/.bashrc, ~/.zshrc, or ~/.bash_profile).\n"
            "   - Add the line: export PATH=/path/to/R:$PATH\n"
            "   - Replace '/path/to/R' with the folder containing the R executable.\n"
            "3. Run `source ~/.bashrc` or restart your terminal."
        )
    else:
        instructions = "Unsupported operating system. Please consult your R installation documentation."
    return instructions


def ensure_r_installed():
    """Ensure R is installed and provide instructions if not."""
    if not is_r_installed():
        print("Error: R is not installed or not found in your system's PATH.")
        print("\nTo install R, visit: https://cran.r-project.org/\n")
        print("After installing R, follow these instructions to add R to your PATH:\n")
        print(get_os_specific_instructions())
        sys.exit(1)  # Exit to prevent further setup steps


if __name__ == "__main__":
    ensure_r_installed()
    install_r_package('quadprog')
    install_r_package('minpack.lm')
    # install_r_package('Rserve') # for Rserver - not in use, it is very slow



# This script not in use - R server is very slow
# R server enable run R with multithreads

import numpy as np
import pyRserve

from AxSI import config


def lin_least_squares_with_constrains_r(A: np.ndarray, b: np.ndarray, lb: np.ndarray, ub: np.ndarray,
                                        start_p: np.ndarray) -> np.ndarray:
    """
    Solve linear least squares with constraints using R's 'quadprog' package via Rserve.

    Parameters:
        A (np.ndarray): The coefficient matrix.
        b (np.ndarray): The target vector.
        lb (np.ndarray): Lower bounds on the solution.
        ub (np.ndarray): Upper bounds on the solution.
        start_p (np.ndarray): Starting point for the linear least squares.

    Returns:
        np.ndarray: The optimized solution vector.
    """
    # Define the D matrix (A^T A) and d vector (-A^T b) for quadprog
    D = A.T @ A
    d = A.T @ b

    # Constraints matrix: combine lb and ub as inequality constraints
    # G matrix: [-I; I] (negative and positive identity matrices for bounds)
    G = np.vstack([np.ones((A.shape[1])), np.eye(A.shape[1]), -np.eye(A.shape[1])])
    h = np.hstack([1, lb, -ub])

    try:
        # Connect to Rserve
        conn = pyRserve.connect(host="localhost", port=config.R_SERVER_PORT)

        # Send the data to R
        conn.assign("D", D.tolist())  # Send as list for R compatibility
        conn.assign("d", d.tolist())
        conn.assign("G", G.tolist())
        conn.assign("h", h.tolist())

        # Run the quadprog solver in R
        conn.void_eval("""
        library(quadprog)
        result <- solve.QP(Dmat = as.matrix(D), dvec = as.numeric(d), 
                           Amat = t(as.matrix(G)), bvec = as.numeric(h), meq = 1)
        solution <- result$solution
        """)

        # Retrieve the solution from R
        solution = np.array(conn.eval("solution"))
        conn.close()  # Close the connection to Rserve

        return solution
    except Exception as e:
        print(f"Error: {e}")
        return np.ones(A.shape[1])  # Return a default solution if the solver fails



import os
import socket
import subprocess
import sys


def find_r_home():
    """Find R_HOME dynamically using the R executable."""
    try:
        # Run R command to get R_HOME
        result = subprocess.run(["R", "RHOME"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True, text=True)
        r_home = result.stdout.strip()
        return r_home
    except FileNotFoundError:
        print("Error: R is not installed or not in the PATH.", file=sys.stderr)
        return None
    except subprocess.CalledProcessError:
        print("Error: Unable to determine R_HOME. Check your R installation.", file=sys.stderr)
        return None


def set_r_environment():
    """Set R_HOME and update PATH based on the detected R installation."""
    r_home = find_r_home()
    if not r_home:
        print("R is not installed or configured properly. Please install R from https://cran.r-project.org/",
              file=sys.stderr)
        sys.exit(1)

    # Set R_HOME environment variable
    os.environ['R_HOME'] = r_home
    print(f"R_HOME set to: {r_home}")

    # Update PATH to include R bin directory
    r_bin_path = os.path.join(r_home, "bin")
    if r_bin_path not in os.environ['PATH']:
        os.environ['PATH'] += os.pathsep + r_bin_path
        print(f"R bin directory added to PATH: {r_bin_path}")

    # Verify R is now accessible
    try:
        subprocess.run(["R", "--version"], check=True)
        print("R is properly configured and accessible.")
    except Exception as e:
        print("Error: R is still not accessible. Ensure it is properly installed.", file=sys.stderr)
        sys.exit(1)


def check_port_availability(port):
    """Check if a port is available (not occupied)."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(('localhost', port))
            return True  # Port is available
        except socket.error:
            return False  # Port is occupied


# Not in use - R server is very slow
def upload_rserver(port=6133):
    if not check_port_availability(port):
        return False
    try:
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.bind(('localhost', port))
        server_socket.listen(5)  # Start listening for connections

        print(f"Server is listening on port {port}...")
        return server_socket
    except Exception as e:
        print(f"Error: {e}")


# Call this function during setup or at runtime as needed
if __name__ == "__main__":
    set_r_environment()
    upload_rserver(int(sys.argv[1]))  # port number (int) - the default is 6133



# This file defines MRI scan object for AxSI analysis
# Scan obect holds various objects for Data, bvals, bvecs and mask

import logging
import os
from pathlib import Path

import nibabel as nb
import numpy as np
from dipy.io import read_bvals_bvecs

from AxSI.matrix_operations import normalize_matrix_rows
from AxSI.toolbox import cart2sph_env, smooth4dim

logger = logging.getLogger(__name__)


class Scan_File:
    """ A basic class holds a file path and its data loaded """

    def __init__(self, filepath: Path, raw_data: np.ndarray) -> None:
        self.filepath = filepath
        self.raw_data = raw_data  # data is already loaded


class NIFTIFile(Scan_File):
    ''' A class holds nifti file such as scan data and mask'''

    def __init__(self, filepath: Path) -> None:
        self.diff_file = nb.load(filepath)
        # load data for parent class
        super().__init__(filepath, self.diff_file.get_fdata())


class Scan:
    """
    Main class for scan object use in the analysis
    holds relevant data from files and parameters
    provides get functions to retreive data
    """

    def __init__(self, file_names: dict, *params) -> None:
        """
        Load data of bvals and bvecs files

        :param file_names: filenames should contain path to four files by specific keys
        :param params:     params should be ordered in specific way
        """
        bvals, bvecs = read_bvals_bvecs(str(file_names['bval']), str(file_names['bvec']))
        self.bval = BVAL(file_names['bval'], bvals)  # init BVAL object
        self.bvec = BVEC(file_names['bvec'], bvecs, self.bval)  # init BVEC object
        self.data = DATA(file_names['data'], self.bval)  # init DATA object
        self.mask = MASK(file_names['mask'])  # init MASK object
        self.shape = self.mask.raw_data.shape  # based on scan resolution. For example: (128, 128, 88) in total 1,441,792
        self.num_of_vols = np.prod(self.shape)  # number of voxels in scan. For example: 342,625
        self.param_dict = self.build_param_dict(*params)

    def build_param_dict(self, small_delta: int, big_delta: int, gmax: float, gamma_val: float = 4257) -> dict:
        """
        create a dictionary with scan parameters

        :param small_delta: (float) Gradient duration in miliseconds.
        :param big_delta: (float) Time to scan (time interval) in milisecond.
        :param gmax: (float) Gradient maximum amplitude in G/cm (or 1/10 mT/m).
                             gmax calculation: sqrt(bval*100/(7.178e8*0.03^2*(0.06-0.01)))
        :param gamma_val: Gyromagnetic ratio.
        :return: dictionary with scan parameters
        """
        grad_dirs = self.bvec.grad_dirs
        # init dictionary
        scan_param = {'nb0': self.bval.locs[0], 'small_delta': small_delta, 'big_delta': big_delta, 'gmax': gmax}
        max_q = gamma_val * small_delta * gmax / 10e6
        # convert grad_dirs to spherical representation
        r_q, phi_q, theta_q = cart2sph_env(grad_dirs.T)  # shape of r_q, phi_q, theta_q: (n_frames,)
        r_q = r_q * max_q  # multiply by coefficient
        scan_param['q_dirs'] = grad_dirs * max_q  # multiply by coefficient
        # set spherical representations
        scan_param['r'] = r_q
        scan_param['phi'] = phi_q
        scan_param['theta'] = theta_q
        return scan_param

    def get_smoothed_data(self, voxel_index=None, bvalue_index=None) -> np.ndarray:
        """
        Get smoothed signal data of mri scan.
        Data is flattened: dim=0 are the voxels, dim=1 time

        Taking smoosed data of specific time (all voxels) or specific voxel (all bvalues)
        If voxel_index=None, bvalue_index=None returns the all smoothed data after masking.

        Note that voxel_index is always None in this program.

        :param voxel_index: for taking data of specific voxels
        :param bvalue_index: for taking data of specific bvalue
        :return: smoothed data
        """
        data = self.data.smoothed[self.mask.index]
        if voxel_index:  # for taking data of specific voxels
            data = data[voxel_index, :]
        if bvalue_index:  # for taking data of specific bvalue
            data = data[:, bvalue_index]
        return data

    def get_signal_0(self, take_first: bool = False) -> np.ndarray:
        """
        :param take_first: (bool) flag for DTI
                            for DTI1000 use only the first appearance of bvalue==0
                            for shellDTI use the mean of all bvalue==0
        :return: 1D array of signal when bvalue==0
        """
        if take_first:
            # first image when bvalue==0
            return self.data.signal_0[self.mask.index]
        else:
            # mean over bvalue==0
            return self.data.signal_0_mean[self.mask.index]

    def get_locs_by_bvalues(self, bvalue) -> list:
        """
        :return a list of indices to images taken with bvalue
        """
        return self.bval.locs[bvalue]

    def get_shape(self) -> np.ndarray:
        """
        :return: brain resolution (of the raw data) (4D)
        """
        return self.shape

    def get_num_of_voxels(self) -> int:
        """
        :return: number of brain voxels (after masking)
                 ** not as the attribute num_of_vol (of raw data) kept in scan **
        """
        return len(self.mask.index)

    def get_params(self) -> dict:
        """
        :return: get dictionary of scan parameters
        """
        return self.param_dict

    def get_bval_data(self) -> np.ndarray:
        """
        :return: array of bval real values
        """
        return self.bval.data * 1000

    def get_bval_length(self) -> int:
        """
        :return: number of timeframes (after filteration)
        """
        return self.bval.length  # n_frames

    def get_shell(self) -> np.ndarray:
        """
        :return: array with unique nonzero elements of bvalues
        """
        return self.bval.shell

    def get_max_bval(self) -> float:
        """
        :return: maximum value of bval
        """
        return self.bval.max_bval

    def get_grad_dirs(self) -> np.ndarray:
        """
        :return: array with bvec grad_dirs data
        """
        return self.bvec.grad_dirs

    def get_bvec_norm(self) -> np.ndarray:
        """
        :return: 2D array with normalized bvec data
        """
        return self.bvec.norm_data

    def reshape_to_brain_size(self, data: np.ndarray) -> np.ndarray:
        """
        :param data: data can be n-dim, but assuming data.shape[0] == len(mask.index), "and `data[0]` contains the
                     data as an array (which may include additional data in other dimensions)"
        :return: matrix of original dimension as raw_data (but in addition can contain additional data in other dimensions)
        """
        exshape = data.shape[1:]  # shape without first dimension, could be empty tuple
        new_data = np.zeros((self.num_of_vols,) + exshape)
        new_data[self.mask.index] = data
        new_data = new_data.reshape(self.shape + exshape)
        return new_data

    def save_files(self, files: dict[str, np.ndarray], subj_folder: Path) -> None:
        """
        Save nifti files in subj_folder based on files dictionary

        :param files: dictionary of files to save. dict should hold strings as keys and np.ndarray as values.
        :param subj_folder: folder to save files
        """
        for key in files.keys():
            filename = build_file_name(subj_folder, key)  # build full path to file
            data = self.reshape_to_brain_size(files[key])  # plant brain data in array of original image dimension
            save_nifti(filename, data, self.data.diff_file.affine)


def save_nifti(fname: str, img: np.ndarray, affine: np.ndarray) -> None:
    """
    Save img as nifti file

    :param fname: full path to save nifti file
    :param img: data to be saved
    :param affine: affine matrix
    """
    file_img = nb.Nifti1Image(img, affine)
    nb.save(file_img, fname)


def build_file_name(subj_folder: Path, file_name: str) -> Path:
    """
    Build path to save files in a specific manner
    """
    return Path(os.path.join(subj_folder, 'AxSI', file_name + '.nii.gz'))


class BVAL(Scan_File):
    """ Object to store data of bvals file """

    def __init__(self, filepath: Path, raw_data: np.ndarray) -> None:
        # call parent constructor with path and raw_data
        super().__init__(filepath, raw_data)
        # map values to rounded integers
        data = self.compute_rounded_data()
        # Remove bval<1000 from calculation
        self.low_locs = self.find_low_locs(data)
        # TODO: (Refael) The deletion of low bvals need to be before the roundness, since after the round there are
        #  or 0 or 1 not between them. But it doesn't affect the results with the new equipment.
        # Remove bval<1000 from calculation
        self.data = np.delete(data, self.low_locs)
        self.length = len(self.data)
        # get unique values of bvalue
        shell = np.unique(self.data)

        self.shell = shell[shell > 0]  # keep only nonzero bvalues
        # store for each bvalue the indices of relevant images
        self.locs = {val: np.where(self.data == val)[0].tolist() for val in shell}
        self.max_bval = np.max(self.data)  # maximal bvalue
        self.norm = np.sqrt(self.data / self.max_bval)  # normalize bvalue data.

    def compute_rounded_data(self) -> np.ndarray:
        """
        Round bvalues to find bad values.
        for example: 1995 => 2, 1500 => 1.5
        :return: rounded_data
        """
        bval2 = 2 * np.asarray(self.raw_data)  # multiply by 2
        rounded_data = bval2.round(-2) / 2000  # round last two digits, divide by 2000
        return rounded_data

    def find_low_locs(self, bval) -> np.ndarray:
        """ get locations where 0 < elements < 1 """
        return np.where((bval > 0) & (bval < 1))[0]


class BVEC(Scan_File):
    """ Object to store data of bvecs file """

    def __init__(self, filepath: Path, raw_data: np.ndarray, bval: BVAL) -> None:
        # call parent constructor with path and loaded data
        super().__init__(filepath, raw_data)
        # use bval to reshape data if needed
        data = np.reshape(self.raw_data, [bval.length, -1])  # dims: (n_frames, 3)
        # use bval to omit bad values
        data = np.delete(data, bval.low_locs, 0)  # data: (n_frames, 3)
        # keep data multiplited by bvals coefficient
        self.grad_dirs = data * bval.norm[np.newaxis].T
        # normalize data
        self.norm_data = normalize_matrix_rows(data)
        self.norm_grad_dirs = normalize_matrix_rows(self.grad_dirs)


class DATA(NIFTIFile):
    """
    Object to store signal data of mri scan.
    Assuming preprocessed for now.
    """

    def __init__(self, filepath: Path, bval: BVAL) -> None:
        # call parent constructor with path
        super().__init__(filepath)
        data = np.asarray(self.raw_data, dtype='float64')
        # use bval to omit bad values
        data = np.delete(data, bval.low_locs, 3)
        data[data <= 0] = 0.0
        X, Y, Z, n = data.shape  # assume data has 4 dimensions
        self.corrected = data
        smoothed_data = smooth4dim(data)  # smooth every image separately
        self.smoothed = smoothed_data.reshape((X * Y * Z, n))  # flatten the brain to make data 2D
        # bval is always 0 in the first time.
        self.signal_0 = self.smoothed[:, bval.locs[0][0]]  # keep the signal when bvalue=0 for the first time
        self.signal_0_mean = np.nanmean(self.smoothed[:, bval.locs[0]], axis=1)  # mean across bvalue=0


class MASK(NIFTIFile):
    """
    Object to store brain mask file
    """

    def __init__(self, filepath) -> None:
        # call parent constructor with path
        super().__init__(filepath)
        self.flattened = self.raw_data.flatten()  # flatten the images to be 1D
        self.index = np.where(self.flattened > 0)[0]  # indices for voxels of brain



# The main file for running AxCaliber analysis
# Use DTI1000, shellDTI and AxSI
import logging
import os
import time
from multiprocessing import Pool
from pathlib import Path

from AxSI import config
from AxSI.ax_dti import dti
from AxSI.calc_axsi import calc_axsi
from AxSI.mri_scan import Scan

logger = logging.getLogger(__name__)


def init_axsi(subj_folder: Path, file_names, *params) -> Scan:
    logger.info("Starting AxSI")
    subf = os.path.join(subj_folder, 'AxSI')
    # make dir subj_folder
    if not os.path.exists(subf):
        os.mkdir(subf)

    scan = Scan(file_names, *params)  # init Scan object
    return scan


def axsi_main(subj_folder: Path, file_names, *params) -> None:
    tic = time.time()
    # initialize analysis
    scan = init_axsi(subj_folder, file_names, *params)  # return Scan object
    toc = time.time()
    logger.debug("Computation time initialization = " + str((toc - tic)) + " sec.")

    shell_dti = []  # array for DTI objects for each bvalue
    bval_shell = scan.get_shell()  # array of unique nonzero bvalues

    logger.info('Starting DTI-1000')
    # run DTI1000
    tic = time.time()
    dti1000 = dti(scan, 1.0, is1000=True, subj_folder=subj_folder)
    toc = time.time()
    logger.debug("Computation time DTI-1000 = " + str((toc - tic)) + " sec.")

    logger.info('Starting DTI-shell')
    # multiprocessing on each voxel is not efficient, instead we run each shell with all voxels in parallel.
    if config.NUM_PROCESSES_AXSI > 1 or config.NUM_PROCESSES_PRED > 1:
        tic = time.time()
        # run parallelly with maximum number of processes that user allocated.
        num_proc = min(max(config.NUM_PROCESSES_PRED, config.NUM_PROCESSES_AXSI), len(bval_shell))
        logger.info(f'Run DTI-shell parallely with {num_proc} processes/threads')
        with Pool(num_proc) as p:  # run volume_dti parallelly
            shell_dti = p.starmap(dti, [(scan, bval_shell[i]) for i in range(len(bval_shell))])
        toc = time.time()
    else:
        for i in range(len(bval_shell)):
            ax_dti = dti(scan, bval_shell[i])  # run DTI
            shell_dti.append(ax_dti)  # store result object in array

    logger.debug("Computation time DTI-shell = " + str((toc - tic)) + " sec.")

    # run AxSI analysis
    logger.info('Starting AxSI')
    calc_axsi(scan, dti1000, shell_dti, subj_folder)



# This file provides methods and basic functionality for AxSI analysis
import logging
import multiprocessing
import os
import time
from multiprocessing import Process, managers
from threading import Thread
from typing import Callable

import gurobipy as gp
import numpy as np
import psutil
from scipy.ndimage import gaussian_filter
from scipy.stats import gamma

from AxSI import config

logger = logging.getLogger(__name__)


def set_log_config():
    # Configure logging
    logging.basicConfig(
        filename=os.path.join(config.SUBJ_FOLDER, 'run_info.log'),  # Log file
        level=logging.DEBUG if config.DEBUG_MODE else logging.INFO,
        # Minimum log level (the available levels: DEBUG, INFO, WARNING, ERROR, CRITICAL)
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'  # Include module name
    )

    # Create a StreamHandler to output to the console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG if config.DEBUG_MODE else logging.INFO)  # Same log level
    console_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

    logging.getLogger().addHandler(console_handler)


def monitor_memory(interval: int = 10) -> None:
    """
    Monitor memory usage over specified interval

    :param interval: monitor interval in seconds
    """
    # Configure logger inside the child process
    # set_log_config()
    # Get the parent process PID
    parent_pid = os.getppid()

    # Get the parent process using its PID
    parent_process = psutil.Process(parent_pid)

    # Monitor memory usage of the parent and its children
    while True:
        total_memory = 0
        try:
            # Add memory usage of the parent process
            total_memory += parent_process.memory_info().rss

            # Iterate over all child processes of the parent
            for child in parent_process.children(recursive=True):
                total_memory += child.memory_info().rss
        except psutil.NoSuchProcess or psutil.AccessDenied or psutil.ZombieProcess:
            continue
        # Convert memory to megabytes (MB)
        total_memory_mb = total_memory / (1024 ** 2)
        logger.debug(f"Total memory usage (Parent + All Children): {total_memory_mb:.2f} MB")

        # Sleep for a while before checking again
        time.sleep(interval)


def parallel_batches(num_items: int, num_par: int, num_thr: int, func: Callable, *args) -> None:
    # Initialize an empty array for processes
    procs = []
    nun_processes = max(num_par, num_thr)
    chunk_size = int(np.ceil(num_items / nun_processes))
    # Calculate the number of chunks (chunk_size voxels per chunk)

    # Create processes for each chunk
    for chunk_idx in range(nun_processes):
        start_idx = chunk_idx * chunk_size
        end_idx = min(start_idx + chunk_size, num_items)  # Ensure the last chunk does not exceed bounds
        if num_thr > 1:
            proc = Thread(
                target=batch_func,
                args=(start_idx, end_idx, func, *args)
            )
        else:
            proc = Process(
                target=batch_func,
                args=(start_idx, end_idx, func, *args)
            )
        procs.append(proc)  # Append process to array
        proc.start()  # Start process

    # Wait for all processes to finish
    for proc in procs:
        proc.join()


def batch_func(start_idx, end_idx, func, *args):
    """
    Processes a batch of tasks by dynamically updating arguments based on index.

    Parameters:
        start_idx (int): Starting index for processing.
        end_idx (int): Ending index for processing.
        func (callable): The function to call with the updated arguments. The last argument must be i.
        *args: The initial arguments to pass to the function.
    """
    if func.__name__ == 'perform_iteration':
        args, gurobi_env = create_new_gurobi_env(args)
    for i in range(start_idx, end_idx):
        func(*args, i)  # Call the function with arguments
    if func.__name__ == 'perform_iteration':
        delete_gurobi_env(gurobi_env)


def create_new_gurobi_env(args):
    # 1) Gurobi does not work correctly with Python's multiprocessing when using the 'fork' start method.
    #    However, our program relies on 'fork' to pass updated configuration values to child processes,
    #    as outlined in the AxSI_main.py script. To address this, here we switch to the 'spawn' method instead.
    #
    # 2) Gurobi is not thread-safe, so a separate environment is required for each process when using
    #    multiprocessing.
    #    For multithreading, it appears to work fine without a separate environment, but we include it as a precaution.

    # See here for more details:
    # https://support.gurobi.com/hc/en-us/articles/360043111231-How-do-I-use-multiprocessing-in-Python-with-Gurobi

    if multiprocessing.get_start_method(allow_none=True) is None:
        multiprocessing.set_start_method("spawn")

    len_args = len(args)
    args = list(args)
    gurobi_env = None
    if config.LINEAR_LSQ_METHOD == 'gurobi':
        gurobi_env = gp.Env()
        if len(args) == len_args:
            args = args + [gurobi_env]
        else:
            args[-1] = gurobi_env
    else:
        args = list(args) + [None]
    return args, gurobi_env


def delete_gurobi_env(gurobi_env):
    del gurobi_env


class Values:
    """
    A basic class to be extended with different attributes.
    Provide get/set methods base on given attribute.
    Used for creating a shared object in parallel prcoessing.
    """

    def get(self, attr, i=-1):
        """
        Get value of given attribute.

        :param attr: assume self uses "attr" to hold array
        :param i:    element i in the array
        :return:     element i if i >= 0, otherwise the whole array
        """
        if i == -1:
            # return array
            return getattr(self, attr)
        # return specific element by index
        return getattr(self, attr)[i]

    def set(self, attr, new_val, i=-1) -> None:
        """
        Set element if i >= 0, otherwise replace the whole array

        :param attr:    assume self uses "attr" to hold array
        :param new_val: new value to be set
        :param i:       element i in the array
        """
        if i == -1:
            # replace array
            setattr(self, attr, new_val)
        else:
            # set specific element by index
            getattr(self, attr)[i] = new_val

    @staticmethod
    def copy_values(source, target, attributes) -> None:
        """
        Copy attributes from source class to target class.
        Assume source and target holds the same attribute.

        :param source: source class
        :param target: target class
        :param attributes: attributes to be copied
        """
        for attr in attributes:
            # replace array in target with source array
            target.set(attr, source.get(attr))

    def get_dict(self, attributes) -> dict:
        """
        Assume self holds data for each attr in attributes
        :param attributes: List of attributes
        :return: dictionary with attributes as keys and data as values
        """
        d = {attr: self.get(attr) for attr in attributes}
        return d

    def nonzero(self, attr) -> None:
        """
        For given attribute in self: set 0 instead of every negative value

        :param attr: attribute to be set 0
        """
        arr = self.get(attr)
        arr[arr <= 0] = 0
        self.set(attr, arr)


# init a new manager for parallel processing 
class MyManager(managers.BaseManager):
    pass


def cart2sph_env(v: np.ndarray) -> tuple:
    """ An envelope function for specific use

    :param v: expected dimension (3, times)
    :return: r, theta, phi
    """
    # Change the sign of the third element
    return cart2sph(v[0], v[1], -v[2])


def cart2sph(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> tuple:
    """
    convert from 3D cartesian representation to spherical representation
    """

    xy = np.sqrt(x ** 2 + y ** 2)  # sqrt(x² + y²)
    r = np.sqrt(xy ** 2 + z ** 2)  # r = sqrt(x² + y² + z²)
    phi = np.arctan2(z, xy)
    theta = np.arctan2(y, x)

    return r, theta, phi


def smooth4dim(data: np.ndarray) -> np.ndarray:
    """
    :param data: 4D array
    :return: 4D array of the same size as input, smooth each 3D slice based on last dimension
    """
    new_data = np.zeros(data.shape)
    for i in range(data.shape[3]):
        new_data[:, :, :, i] = smooth_data_slice(data[:, :, :, i])
    return new_data


def smooth_data_slice(slice: np.ndarray) -> np.ndarray:
    """
    Use gaussian filter to smooth input data with pre-defined parameters
    :param slice: 3D array
    :return: 3D array of the same size as input, smooth each 3D slice based on last dimension
    """
    smoothed_slice = gaussian_filter(slice, sigma=0.65, truncate=3, mode='nearest')
    return smoothed_slice


def init_yd() -> np.ndarray:
    """
    Initialization of array for calc_axsi

    :return: yd.shape == ADD_VALS.shape
    """
    alpha = 3
    beta = 2
    gamma_pdf = gamma.pdf(config.ADD_VALS, a=alpha, scale=beta)
    yd = gamma_pdf * np.pi * (config.ADD_VALS / 2) ** 2
    yd = yd / np.sum(yd)
    return yd


def init_l_matrix(n) -> np.ndarray:
    """
    Initialization of array for calc_axsi

    :param n: N_SPEC (len(ADD_VALS))
    :return:  l_mat.shape == (n, n+2) = (N_SPEC, N_SPEC+2)
              for n == 3 output is:
                         1, 0, 0, 0, 0
                        -1, 1, 0, 0, 0
                         0,-1, 1, 0, 0
    """
    ones = np.ones(n)
    # create ID matrix with a diagonal of -1 below the main diag
    l_mat = np.eye(n) - np.tril(ones, -1) * np.triu(ones, -1)
    # add two columns of 0
    l_mat = np.append(l_mat, np.zeros((n, 2)), axis=1)
    return l_mat



import logging
from typing import Optional, Callable, Tuple, Any

import numpy as np
import rpy2.robjects as ro
from lsqAxSI.lsq_components import LSQResult, defaultopt
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FloatVector

minpack = importr('minpack.lm')

logger = logging.getLogger(__name__)


def reg_func(x: np.ndarray, ydata: np.ndarray, pixpredictH: np.ndarray, pixpredictR: np.ndarray,
             pixpredictCSF: np.ndarray, prcsf: float) -> np.ndarray:
    """
    Regression function for nonlinear least squares.

    Computes the error between observed data (`ydata`) and predicted data based on
    a model combining hindered, restricted, and CSF signal components, using a set
    of parameters.

    :param x: Array of parameters for the model, where:
              - x[0]: Fraction of the hindered component.
              - x[1]: Scaling factor for the predicted signal.
    :param ydata: Observed signal data. Shape: (n_frames,).
    :param pixpredictH: Predicted hindered component of the signal. Shape: (n_frames,).
    :param pixpredictR: Predicted restricted component of the signal. Shape: (n_frames,).
    :param pixpredictCSF: Predicted CSF (cerebrospinal fluid) component of the signal. Shape: (n_frames,).
    :param prcsf: Fraction of the CSF component in the signal.
    :return: The residual error between the observed (`ydata`) and the predicted signal.
             Shape: (n_frames,).
    """
    # xt = 1 - x[0] - prcsf
    # newdata = x[1] * (x[0] * pixpredictH + xt * pixpredictR + prcsf * pixpredictCSF)
    newdata = x[1] * (x[0] * pixpredictH + (1 - x[0] - prcsf) * pixpredictR + prcsf * pixpredictCSF)
    err = newdata - ydata
    return err


def jac_calc(x, ydata, pixpredictH, pixpredictR, pixpredictCSF, prcsf) -> np.ndarray:
    """ jacobian matrix calculation for nonlinear least squares (reg_func function). dim: (n_frames, 2) """
    jac = np.zeros([len(ydata), 2])
    jac[:, 0] = x[1] * (pixpredictH - pixpredictR)
    jac[:, 1] = x[0] * pixpredictH + (1 - x[0] - prcsf) * pixpredictR + prcsf * pixpredictCSF
    return jac


# Define R function in Python
reg_func_r = ro.r('''
function(x, ydata, pixpredictH, pixpredictR, pixpredictCSF, prcsf) {
    xt <- 1 - x[1] - prcsf
    newdata <- x[2] * (x[1] * pixpredictH + xt * pixpredictR + prcsf * pixpredictCSF)
    err <- newdata - ydata
    return(err)
}
''')

jac_calc_r = ro.r('''
function(x, ydata, pixpredictH, pixpredictR, pixpredictCSF, prcsf) {
  # Initialize Jacobian matrix
  n_frames <- length(ydata)
  jac <- matrix(0, nrow = n_frames, ncol = 2)

  # Calculate the first column
  jac[, 1] <- x[2] * (pixpredictH - pixpredictR)

  # Calculate the second column
  jac[, 2] <- x[1] * pixpredictH + (1 - x[1] - prcsf) * pixpredictR + prcsf * pixpredictCSF

  return(jac)
}

''')


def least_squares_envelope_r(reg_func: Callable, x0: np.ndarray, bounds: Tuple[np.ndarray, np.ndarray], jac: Callable,
                             ftol: float = 1e-6, xtol: float = 1e-6, diff_step: float = 1e-3,
                             max_nfev: int = 20000, args: Optional[Tuple[Any, ...]] = None) -> LSQResult:
    x0_r = FloatVector(x0)
    min_val_r = FloatVector(bounds[0])
    max_val_r = FloatVector(bounds[1])
    ydata_r = FloatVector(args[0])
    pixpredictH_r = FloatVector(args[1])
    pixpredictR_r = FloatVector(args[2])
    pixpredictCSF_r = FloatVector(args[3])
    prcsf_r = float(args[4])

    # Define control settings
    control = minpack.nls_lm_control(maxiter=defaultopt['MaxIter'], maxfev=max_nfev, ftol=ftol)

    # Run the optimization
    result = minpack.nls_lm(
        par=x0_r,
        fn=reg_func,
        jac=jac,
        lower=min_val_r,
        upper=max_val_r,
        control=control,
        ydata=ydata_r,
        pixpredictH=pixpredictH_r,
        pixpredictR=pixpredictR_r,
        pixpredictCSF=pixpredictCSF_r,
        prcsf=prcsf_r
    )
    # Extract and return the optimized parameters
    return dict(result.items())['par']
